import c4d
from c4d import gui
import json, sys, socket, time, os, random, time
from c4d import storage
from c4d.threading import C4DThread
from ctypes import pythonapi, c_void_p, py_object

#Global server reference.
thread_ = None
host, port = '127.0.0.1', 24420
#current plugin version
pluginVersion = "2.0"
#This plugin ID was acquired from Maxon's PluginCafe Forum
pluginID = 1051284
pluginIDCommandData = 1051282
BC_WORLD_ID = 1051281
BC_IMPORT_MS_ASSET = 1051285
BC_ADD_OBJ = 1051283
BC_DO_SOMETHING_ELSE = 1000001

#Background_Server is driven from Thread class in order to make it run in the background.
class BGThread(C4DThread):
    end = False

    # Called by TestBreak to adds a custom condition to leave
    def TestDBreak(self):
        return bool(self.end) 

    def GetFreeBcId(self, bc):
        # If bc is empty we will use the first entry
        if not len(bc):
            return 0
        
        # Get all index used
        usedIds = [x for x, y in bc if y is not None]
        
        # If the min id used is superior than 0, means 0 is free, we use it
        if min(usedIds) > 0:
            return 0
        
        # We iterate beetwen 0 and max to see if there is an empty ID
        if(c4d.GetC4DVersion() < 23000):
            for i in xrange(0, max(usedIds)):
                if i not in usedIds:
                    return i
        else:
            for i in range(0, max(usedIds)):
                if i not in usedIds:
                    return i
        
        # And finally if there is no "hole" in the list we add it
        return len(usedIds)
    
    #Start the thread to start listing to the port.
    def Main(self):

        try:
            self.socket_ = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket_.bind((host, port))

            while True:
                # Check if the thread need to be closed e.g. a call of thread_.End()
                if self.TestBreak():
                    return

                time.sleep(0.01)

                self.socket_.listen(5)
                client, addr = self.socket_.accept()
                data = ""
                buffer_size = 4096*2
                data = client.recv(buffer_size)

                # Check if the thread need to be closed e.g. a call of thread_.End()
                if self.TestBreak():
                    return
                elif data == "Bye Megascans":
                    self.end = True
                    self.End()
                    return

                if data != "":
                    self.TotalData = b""
                    self.TotalData += data
                    while True:
                        # Check if the thread need to be closed e.g. a call of thread_.End()
                        if self.TestBreak():
                            return
                        data = client.recv(4096*2)
                        
                        # Check if the thread need to be closed e.g. a call of thread_.End()
                        if self.TestBreak():
                            return
                        elif data == "Bye Megascans":
                            self.end = True
                            self.End()
                            return

                        if data : self.TotalData += data
                        else:
                            if(self.TotalData != None):
                                
                                # Get the world container
                                bcWorld = c4d.GetWorldContainerInstance()

                                # Get the basecontainer which will contain all our Message
                                bcPlugin = bcWorld.GetContainerInstance(BC_WORLD_ID)
                                if not bcPlugin:
                                    bcPlugin = c4d.BaseContainer()
                                    bcWorld[BC_WORLD_ID] = bcPlugin
                                    bcPlugin = bcWorld.GetContainerInstance(BC_WORLD_ID)

                                # Serialize the data into a BaseContainer, just for the demonstration purpose I handle 2 kind of data to demonstrate how to identify them later
                                bcEvent = c4d.BaseContainer(BC_IMPORT_MS_ASSET)
                                if(c4d.GetC4DVersion() < 23000):
                                    bcEvent[0] = self.TotalData
                                else:
                                    bcEvent[0] = self.TotalData.decode('utf-8')

                                # Get a freeItem in our plugin basecontainer, stored in the world container
                                index = 0#self.GetFreeBcId(bcPlugin)
                                bcPlugin[index] = bcEvent

                                # Now that our data is in the world container we send the message to trigger the MessageData
                                c4d.SpecialEventAdd(pluginID, int(index), 0)

                                self.TotalData = None
                            break
        except Exception as e:
            print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)

class MS_Importer ():

    def __init__(self, imported_data, useEXRs, applyMatToSelection, useDisplacement_):
        if(c4d.GetC4DVersion() < 19000):
            show_error_message("Your current version of C4D is not supported. Supported versions of Cinema4D are: R19 and R20")
            return

        self.renderEngine = getRenderEngine()
        if(self.renderEngine != 0 and  self.renderEngine != 1036219 and  self.renderEngine != 1029988 and  self.renderEngine != 1029525): 
            show_error_message("Your current render engine is not supported. Supported render engines are: Physical, Standard, Octane, Redshift, and Arnold")
            return

        try:
            self.current_geo_import = []
            self.useDisplacement = useDisplacement_
            json_array = json.loads(imported_data)

            for js in json_array:

                self.json_data = js

                self.selectedObjects = []
                print("Quixel Bridge Plugin::Importing " + self.json_data["name"])
                self.assetType = self.json_data["type"]
                self.assetPath = self.json_data["path"]
                self.assetID = self.json_data["id"]
                self.previewSize = 11
                self.isHighPoly = bool(self.json_data["activeLOD"] == "high")
                self.activeLOD = self.json_data["activeLOD"]
                self.minLOD = self.json_data["minLOD"]
                self.isMetal = bool(self.json_data["category"] == "Metal")
                self.hasVariations = False
                self.isScatterAsset = self.CheckScatterAsset(self.json_data)
                self.isSpecularWorkflow = True
                self.isBillboard = self.CheckIsBillboard()
                self.isScatterObj = bool(self.isScatterAsset and "obj" in self.json_data["meshFormat"])
                self.isMeshObj = bool("obj" in self.json_data["meshFormat"])
                self.applyToSelection = applyMatToSelection and bool(self.assetType == "surface")
                self.isOpaque = True
                self.isScaleFixed = False

                if "workflow" in self.json_data.keys():
                    self.isSpecularWorkflow = bool(self.json_data["workflow"] == "specular")

                if self.json_data["resolution"] == "4K":
                    self.previewSize = 12
                elif self.json_data["resolution"] == "8K":
                    self.previewSize = 13

                texturesListName = "components" if self.isBillboard else "components"

                self.available_textures = [obj["type"] for obj in self.json_data[texturesListName]]
                self.texture_list = []

                for obj in self.json_data[texturesListName]:
                        texFormat = obj["format"].encode('utf-8')
                        texType = obj["type"].encode('utf-8')
                        texPath = obj["path"].encode('utf-8')
                        
                        # Replace diffuse texture type with albedo so we don't have to add more conditions to handle diffuse map.
                        if texType == "diffuse" and "albedo" not in self.available_textures:
                            texType = "albedo"
                            self.available_textures.append("albedo")
                            self.available_textures.remove("diffuse")

                        if texType == "displacement":
                            self.available_textures.remove("displacement")
                            if getRenderEngine() == 1029525 and useDisplacement_: #Octane displacement should be JPG only.
                                if texFormat.lower() == "jpg":
                                    self.texture_list.append((texFormat, texType, texPath))
                                    self.available_textures.append("displacement")
                                else:
                                    texDir = os.path.dirname(texPath)
                                    texName = os.path.splitext(os.path.basename(texPath))[0]
                                    if os.path.exists(os.path.join(texDir, texName + ".jpg")):
                                        texPath = os.path.join(texDir, texName + ".jpg")
                                        texFormat = "jpg"
                                        self.texture_list.append((texFormat, texType, texPath))
                                        self.available_textures.append("displacement")
                            elif useDisplacement_: # Other render engines.
                                if useEXRs:
                                    texDir = os.path.dirname(texPath)
                                    texName = os.path.splitext(os.path.basename(texPath))[0]
                                    if os.path.exists(os.path.join(texDir, texName + ".exr")):
                                        texPath = os.path.join(texDir, texName + ".exr")
                                        texFormat = "exr"
                                        self.texture_list.append((texFormat, texType, texPath))
                                        self.available_textures.append("displacement")
                                else:
                                    self.texture_list.append((texFormat, texType, texPath))
                                    self.available_textures.append("displacement")
                        else:
                            self.texture_list.append((texFormat, texType, texPath))

                self.geometryList = [(obj["format"].encode('utf-8'), obj["path"].encode('utf-8')) for obj in self.json_data["meshList"]]

                if "name" in self.json_data.keys():
                    self.assetName = self.json_data["name"].replace(" ", "_")
                else:
                    self.assetName = os.path.basename(self.json_data["path"]).replace(" ", "_")
                if len(self.assetName.split("_")) > 2:
                    self.assetName = "_".join(self.assetName.split("_")[:-1])

                self.materialName = self.assetName + '_' + self.assetID

                if self.assetType == "brush":
                    print("Quixel Bridge Plugin::Asset type: Brush is not supported in C4D yet.")
                    return
                
                # Scale fixed assets check
                for meta in self.json_data["meta"]:
                    if meta['key'] == 'isScaleFixed' and meta['value'].lower() == 'true':
                        self.isScaleFixed = True
                    if meta['key'] == 'small3dasset' and meta['value'].lower() == 'true':
                        self.isScaleFixed = True

                self.initImportProcess()
                print("Quixel Bridge Plugin::Successfully imported " + self.json_data["name"])

        except Exception as e:
            error_msg = "Message: Failed to read the asset data. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
            show_error_message(error_msg)
            print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)

    def initImportProcess (self):
        try:
            self.imported_geo = []
            self.existing_geo = []
            doc = c4d.documents.GetActiveDocument()
            self.has_geometry = False 
            
            self.existing_geo = doc.GetObjects()
            self.has_geometry = False
            if len(self.geometryList) >= 1:
                self.has_geometry = True
                self.import_geometry()
                current_obj = doc.GetObjects()
                for i in current_obj:
                    if i not in self.existing_geo:
                        if not self.isScaleFixed:
                            #Scaling geometry to 2.54 for accurate displacement
                            i[c4d.ID_BASEOBJECT_REL_SCALE,c4d.VECTOR_X] = 2.54
                            i[c4d.ID_BASEOBJECT_REL_SCALE,c4d.VECTOR_Y] = 2.54
                            i[c4d.ID_BASEOBJECT_REL_SCALE,c4d.VECTOR_Z] = 2.54
                        self.imported_geo.append(i)
            
            if self.isScatterObj and len(self.imported_geo) > 0: # Scatter assets in obj format are imported differently. So we handle them separately.
                self.imported_geo[0].SetName(self.assetID + '_' +self.assetName)
                childObjs = self.imported_geo[0].GetChildren()
                self.imported_geo = childObjs
            elif(self.isScatterAsset and len(self.imported_geo) >= 2):
                scatterParent = c4d.BaseObject(c4d.Onull)   
                scatterParent.SetName(self.assetID + '_' +self.assetName)   
                # Insert scatter parent into scene
                doc.InsertObject(scatterParent)
                # parent geo to the scatter parent object.
                for obj in self.imported_geo:
                    obj.InsertUnder(scatterParent)
            elif self.isMeshObj and len(self.imported_geo) == 1: # Modular assets not tagged as scatter.
                if len(self.imported_geo[0].GetChildren()) > 0:
                    self.imported_geo[0].SetName(self.assetID + '_' +self.assetName)
                    childObjs = self.imported_geo[0].GetChildren()
                    self.imported_geo = childObjs
            
            if len(self.texture_list) >= 1:
                self.setup_material()

            Refresh()
        except Exception as e:
            error_msg = "Message: Failed to process the asset data. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
            show_error_message(error_msg)
            print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)

    def import_geometry (self):
        try:
            if len(self.geometryList) >= 1:
                for obj in self.geometryList:
                    if(c4d.GetC4DVersion() < 23000):
                        AddGeometry(obj[1])
                    else:
                        AddGeometry(obj[1].decode('utf-8'))
                    
        except Exception as e:
            error_msg = "Message: Failed to import geometry. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
            show_error_message(error_msg)
            print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)

    def CheckScatterAsset(self, assetJson):
        if('scatter' in assetJson['categories'] or 'scatter' in assetJson['tags'] or 'cmb_asset' in assetJson['categories'] or 'cmb_asset' in assetJson['tags']):
            return True
        return False

    def CheckIsBillboard(self):
        # Use billboard textures if importing the Billboard LOD.
        if(self.assetType == "3dplant"):
            if (self.activeLOD == self.minLOD):
                return True
        return False

    #Material setup according to the current Render Engine selected in render settings.
    def setup_material (self):
        try:
            ''' Render Engine Ids
            ("ProRender", 1037639),
            ("Arnold", 1029988),
            ("Corona", 1030480),
            ("Octane", 1029525),
            ("Physical", 0),
            ("Redshift", 1036219),
            ("V-Ray", 1019782)]
            '''

            if not (self.renderEngine == 0 or self.renderEngine == 1036219 or self.renderEngine == 1029988 or self.renderEngine == 1029525):
                errorMessage = 'Your current render engine is not supported by the Bridge Plugin so we are terminating the import process but the Plugin is still running!'
                show_error_message(errorMessage)

            mat = None
            if(self.renderEngine == 0): # Physical/Standard
                mat = physical_material(self.texture_list, self.available_textures, self.materialName, self.has_geometry, self.previewSize, self.isMetal)
                if mat != None:
                    ApplyMaterialToGeometry(mat, self.imported_geo, self.applyToSelection, self.isOpaque, self.useDisplacement, self.current_geo_import)
            elif(self.renderEngine == 1036219): # Redshift
                mat = redshift_material(self.texture_list, self.available_textures, self.materialName, self.isMetal)
                if mat != None:
                    ApplyMaterialToGeometry(mat, self.imported_geo, self.applyToSelection, self.isOpaque, self.useDisplacement, self.current_geo_import)
            elif(self.renderEngine == 1029988): # Arnold
                mat = arnold_material(self.texture_list, self.available_textures, self.materialName, self.has_geometry, self.isMetal)
                if mat != None:
                    ApplyMaterialToGeometry(mat, self.imported_geo, self.applyToSelection, self.isOpaque, self.useDisplacement, self.current_geo_import)
            elif(self.renderEngine == 1029525): # Octane
                octaneHelper = OctaneSetup(self.texture_list, self.available_textures, self.assetType, self.has_geometry, self.materialName, self.isSpecularWorkflow, self.isMetal)
                mat = octaneHelper.CreateMaterial()
                if mat != None:
                    ApplyMaterialToGeometry(mat, self.imported_geo, self.applyToSelection, self.isOpaque, self.useDisplacement, self.current_geo_import)
        except Exception as e:
            error_msg = "Message: Failed to setup material. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
            show_error_message(error_msg)
            print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)

############# PYHSICAL SETUP #########################
def physical_material (texture_list, available_textures, matName, has_geometry, previewSize, isMetal):
    try:
        mat = c4d.Material()
    except Exception as e:
        error_msg = "Message: Failed to create the material. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
        show_error_message(error_msg)
        print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)
        return None

    mat.SetName(matName)

    mat.RemoveReflectionLayerIndex(0) #Remove default specular layer
    mat.AddReflectionLayer() # Add lambertian diffuse layer
    mat.AddReflectionLayer() # Add GGX layer
    
    mat[c4d.MATERIAL_USE_COLOR] = True

    color_layer = mat.GetReflectionLayerIndex(1) # Get color layer reference
    refl_layer = mat.GetReflectionLayerIndex(0) # Get metal/non-metal layer reference

    color_layer.SetName("Base Color")
    refl_layer.SetName("Metal" if isMetal else "Non-Metal")
    # Setting layers types to lambertian and GGX
    mat[color_layer.GetDataID() + c4d.REFLECTION_LAYER_MAIN_DISTRIBUTION] = 7 
    mat[color_layer.GetDataID() + c4d.REFLECTION_LAYER_MAIN_VALUE_SPECULAR] = 0 # set Specular strength
    mat[color_layer.GetDataID() + c4d.REFLECTION_LAYER_MAIN_VALUE_BUMP_MODE] = 2 # Set Bump strength mode

    mat[refl_layer.GetDataID() + c4d.REFLECTION_LAYER_MAIN_DISTRIBUTION] = 3
    mat[refl_layer.GetDataID() + c4d.REFLECTION_LAYER_MAIN_VALUE_SPECULAR] = 0 # set Specular strength
    mat[refl_layer.GetDataID() + c4d.REFLECTION_LAYER_MAIN_VALUE_BUMP_MODE] = 2 # Set Bump strength mode
    mat[refl_layer.GetDataID() + c4d.REFLECTION_LAYER_MAIN_VALUE_ROUGHNESS] = 1 # 100% roughness
    # Set layer Fresnel properties according to type.
    if isMetal:
        mat[refl_layer.GetDataID() + c4d.REFLECTION_LAYER_MAIN_VALUE_SPECULAR] = 1 # 100% specular strength
        mat[refl_layer.GetDataID() + c4d.REFLECTION_LAYER_FRESNEL_MODE] = 2 #conductor
        mat[refl_layer.GetDataID() + c4d.REFLECTION_LAYER_FRESNEL_VALUE_ETA] = 1.52 #IOR
        mat[refl_layer.GetDataID() + c4d.REFLECTION_LAYER_FRESNEL_OPAQUE] = True # Opaque
    else:
        mat[refl_layer.GetDataID() + c4d.REFLECTION_LAYER_FRESNEL_MODE] = 1 #dielectric
        mat[refl_layer.GetDataID() + c4d.REFLECTION_LAYER_FRESNEL_VALUE_IOR] = 1.52 #IOR
    
    try:
        
        for item in texture_list:
            if(c4d.GetC4DVersion() >= 23000):
                item = (item[0].decode('utf-8'), item[1].decode('utf-8'), item[2].decode('utf-8'))
            if item[1] == "albedo":
                create_bitmap_shader(mat, item[2], "Albedo", color_layer.GetDataID() + c4d.REFLECTION_LAYER_COLOR_TEXTURE, "srgb")
                create_bitmap_shader(mat, item[2], "Albedo", c4d.MATERIAL_COLOR_SHADER, "srgb")
            elif item[1] == "metalness":
                create_bitmap_shader(mat, item[2], "Metalness", refl_layer.GetDataID() + c4d.REFLECTION_LAYER_TRANS_TEXTURE, "linear", not isMetal)
            elif item[1] == "roughness":
                create_bitmap_shader(mat, item[2], "Roughness", refl_layer.GetDataID() + c4d.REFLECTION_LAYER_MAIN_SHADER_ROUGHNESS, "linear")
            elif item[1] == "gloss" and "roughness" not in available_textures:
                create_bitmap_shader(mat, item[2], "Gloss", refl_layer.GetDataID() + c4d.REFLECTION_LAYER_MAIN_SHADER_ROUGHNESS, "linear", True)
            elif item[1] == "normal":
                create_bitmap_shader(mat, item[2], "Normals", color_layer.GetDataID() + c4d.REFLECTION_LAYER_MAIN_SHADER_BUMP_CUSTOM, "linear")
                create_bitmap_shader(mat, item[2], "Normals", refl_layer.GetDataID() + c4d.REFLECTION_LAYER_MAIN_SHADER_BUMP_CUSTOM, "linear")
            elif item[1] == "displacement":
                mat[c4d.MATERIAL_USE_DISPLACEMENT] = True
                create_bitmap_shader(mat, item[2], "Displacement", c4d.MATERIAL_DISPLACEMENT_SHADER, "linear")
                mat[c4d.MATERIAL_DISPLACEMENT_HEIGHT] = 1 if has_geometry else 10
                mat[c4d.MATERIAL_DISPLACEMENT_SUBPOLY] = True
            elif item[1] == "opacity":
                mat[c4d.MATERIAL_USE_ALPHA] = True
                create_bitmap_shader(mat, item[2], "Opacity", c4d.MATERIAL_ALPHA_SHADER, "linear")
            elif item[1] == "translucency":
                mat[c4d.MATERIAL_USE_LUMINANCE] = True
                sssShader = create_sss_shader(mat, c4d.MATERIAL_LUMINANCE_SHADER)
                create_bitmap_shader(sssShader, item[2], "Translucency", c4d.XMBSUBSURFACESHADER_SHADER, "srgb")
            elif item[1] == "transmission" and "translucency" not in available_textures:
                mat[c4d.MATERIAL_USE_LUMINANCE] = True
                sssShader = create_sss_shader(mat, c4d.MATERIAL_LUMINANCE_SHADER)
                create_bitmap_shader(sssShader, item[2], "Transmission", c4d.XMBSUBSURFACESHADER_SHADER, "linear")

    except Exception as e:
        error_msg = "Message: Failed to import the textures. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
        show_error_message(error_msg)
        print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)
        return None

    mat[c4d.MATERIAL_PREVIEWSIZE] = previewSize
    mat.Message(c4d.MSG_UPDATE)
    mat.Update(True, True)
    doc = c4d.documents.GetActiveDocument()
    doc.InsertMaterial(mat)
    return mat

# Create a bitmap shader to load a texture
def create_bitmap_shader(mat, texPath, name, link_name, colorspace, invert = False):
    shader = c4d.BaseList2D(c4d.Xbitmap)
    shader[c4d.BITMAPSHADER_FILENAME] = texPath
    if invert:
        shader[c4d.BITMAPSHADER_BLACKPOINT] = 1
        shader[c4d.BITMAPSHADER_WHITEPOINT] = 0
    shader[c4d.BITMAPSHADER_COLORPROFILE] = 1 if colorspace == "linear" else 2
    shader.SetName(name)
    mat[link_name] = shader
    mat.InsertShader(shader)
    return shader

# Create a sss shader
def create_sss_shader(mat, link_name):
    shader = c4d.BaseList2D(1025614)
    shader[c4d.XMBSUBSURFACESHADER_DIFFUSE] = c4d.Vector(1,1,1)
    mat[link_name] = shader
    mat.InsertShader(shader)
    return shader

############# REDSHIFT SETUP #########################
def redshift_material (texture_list, available_textures, matName, isMetal):
    try:
        import redshift # Try tp import the Redshift API.
        doc = c4d.documents.GetActiveDocument()
        c4d.CallCommand(1036759, 1000)
        mat = doc.GetFirstMaterial()
        mat.SetName(matName)
        node_master = redshift.GetRSMaterialNodeMaster(mat)
        output = node_master.GetRoot().GetDown()

        rs_material = output.GetNext()
        rs_material[c4d.REDSHIFT_SHADER_MATERIAL_REFL_BRDF] = 1
        rs_material[c4d.REDSHIFT_SHADER_MATERIAL_REFL_FRESNEL_MODE] = 2
        rs_material[c4d.REDSHIFT_SHADER_MATERIAL_REFL_METALNESS] = 1 if isMetal else 0
        rs_material.AddPort(c4d.GV_PORT_INPUT, c4d.DescID(c4d.DescLevel(c4d.REDSHIFT_SHADER_MATERIAL_DIFFUSE_COLOR)), message=True)
        rs_material.AddPort(c4d.GV_PORT_INPUT, c4d.DescID(c4d.DescLevel(c4d.REDSHIFT_SHADER_MATERIAL_REFL_ROUGHNESS)), message=True)

        normal_index = 2 # normal map input slot#
        if "metalness" in available_textures:
            rs_material.AddPort(c4d.GV_PORT_INPUT, c4d.DescID(c4d.DescLevel(c4d.REDSHIFT_SHADER_MATERIAL_REFL_METALNESS)), message=True)
            normal_index = 3
        
        
        translucency_index = normal_index + 1 # translucency map input slot#
        transmission_index = normal_index + 1 # transmission map input slot#

        rs_material.AddPort(c4d.GV_PORT_INPUT, c4d.DescID(c4d.DescLevel(10062)), message=True)

        if "translucency" in available_textures:
            transmission_index = 5
            rs_material.AddPort(c4d.GV_PORT_INPUT, c4d.DescID(c4d.DescLevel(c4d.REDSHIFT_SHADER_MATERIAL_TRANSL_COLOR)), message=True)
        if "transmission" in available_textures:
            rs_material.AddPort(c4d.GV_PORT_INPUT, c4d.DescID(c4d.DescLevel(c4d.REDSHIFT_SHADER_MATERIAL_TRANSL_WEIGHT)), message=True)

    except Exception as e:
        error_msg = "Message: Failed to create redshift material. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
        show_error_message(error_msg)
        print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)
        return None
    try:
        for item in texture_list:
            if(c4d.GetC4DVersion() >= 23000):
                item = (item[0].decode('utf-8'), item[1].decode('utf-8'), item[2].decode('utf-8'))
            if item[1] == "albedo":
                albedo_node = create_rs_tex_node(node_master, item[2], "Albedo", 0, rs_material, -100, 200, ".exr" not in item[2])
                albedo_node.GetOutPort(0).Connect(rs_material.GetInPort(0))
                if "ao" in available_textures:
                    for tex in texture_list:
                        if tex[1] == "ao":
                            albedo_node.AddPort(c4d.GV_PORT_INPUT, c4d.DescID(c4d.DescLevel(c4d.REDSHIFT_SHADER_TEXTURESAMPLER_COLOR_MULTIPLIER)), message=True)
                            create_rs_tex_node(node_master, tex[2], "AO", 0, albedo_node, -250, 200)

            elif item[1] == "gloss" and "roughness" not in available_textures:
                gloss_invert_node = create_rs_tex_node(node_master, item[2], "Gloss", 1, rs_material, -100, 250, False, True)
                gloss_invert_node.GetOutPort(0).Connect(rs_material.GetInPort(1))
            elif item[1] == "roughness":
                create_rs_tex_node(node_master, item[2], "Roughness", 1, rs_material, -100, 250)
            elif item[1] == "metalness":
                create_rs_tex_node(node_master, item[2], "Metalness", 2, rs_material, -100, 300)
            elif item[1] == "normal":
                rs_bump_node = create_redshift_base_node(node_master, "BumpMap", "BumpMap", -100, 350)
                rs_bump_node.AddPort(c4d.GV_PORT_INPUT, c4d.DescID(c4d.DescLevel(10000)), message=True)
                rs_bump_node[c4d.REDSHIFT_SHADER_BUMPMAP_INPUTTYPE] = 1
                create_rs_tex_node(node_master, item[2], "Normal", 0, rs_bump_node, -250, 350)
                rs_bump_node.GetOutPort(0).Connect(rs_material.GetInPort(normal_index))
            elif item[1] == "opacity":
                rs_sprite_material = create_redshift_base_node(node_master, "Sprite", "Opacity", 200, 200)
                rs_sprite_material[c4d.REDSHIFT_SHADER_SPRITE_TEX0_GAMMAOVERRIDE] = True
                rs_sprite_material[c4d.REDSHIFT_SHADER_SPRITE_TEX0, c4d.REDSHIFT_FILE_PATH] = item[2]
                rs_sprite_material.GetOutPort(0).Connect(output.GetInPort(0))
                rs_sprite_material.AddPort(c4d.GV_PORT_INPUT, c4d.DescID(c4d.DescLevel(10000)), message=True)
                rs_material.GetOutPort(0).Connect(rs_sprite_material.GetInPort(0))
            elif item[1] == "translucency":
                rs_material[c4d.REDSHIFT_SHADER_MATERIAL_TRANSL_WEIGHT] = 1
                create_rs_tex_node(node_master, item[2], "Translucency", translucency_index, rs_material, -100, 450, True)
            elif item[1] == "transmission":
                rs_material[c4d.REDSHIFT_SHADER_MATERIAL_REFR_WEIGHT] = 0.1
                rs_material[c4d.REDSHIFT_SHADER_MATERIAL_REFR_ABSORPTION_SCALE] = 0.1
                create_rs_tex_node(node_master, item[2], "Transmission", transmission_index, rs_material, -100, 500)
            elif item[1] == "displacement":
                output.AddPort(c4d.GV_PORT_INPUT, c4d.DescID(c4d.DescLevel(10001)), message=True)
                rs_disp_node = create_redshift_base_node(node_master, "Displacement", "RS Displacement", 100, 500)
                rs_disp_node.AddPort(c4d.GV_PORT_INPUT, c4d.DescID(c4d.DescLevel(10000)), message=True)
                rs_disp_node[c4d.REDSHIFT_SHADER_DISPLACEMENT_SPACE_TYPE] = 2
                disp_tex_node = create_rs_tex_node(node_master, item[2], "Displacement", 0, rs_disp_node, False, -100, 550)
                disp_tex_node[c4d.REDSHIFT_SHADER_TEXTURESAMPLER_TEX0_GAMMAOVERRIDE] = True
                rs_disp_node.GetOutPort(0).Connect(output.GetInPort(1))

        mat.Update(True, True)
        return mat

    except Exception as e:
        error_msg = "Message: Failed to import the textures. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
        show_error_message(error_msg)
        print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)
        return None

def create_redshift_base_node(node_master, nodetype, name,  posX, posY):
    rs_node = node_master.CreateNode(node_master.GetRoot(), 1036227, None, posX, posY)
    rs_node[c4d.GV_REDSHIFT_SHADER_META_CLASSNAME] = nodetype
    rs_node[c4d.ID_BASELIST_NAME] = name
    return rs_node

def create_rs_tex_node(node_master, texPath, name, link, rs_material, posX, posY, srgb = False, invert = False):
    rs_tex_node = create_redshift_base_node(node_master, "TextureSampler", name, posX, posY)
    rs_tex_node[c4d.REDSHIFT_SHADER_TEXTURESAMPLER_TEX0, c4d.REDSHIFT_FILE_PATH] = texPath
    rs_tex_node[c4d.REDSHIFT_SHADER_TEXTURESAMPLER_TEX0_GAMMAOVERRIDE] = True
    rs_tex_node[c4d.REDSHIFT_SHADER_TEXTURESAMPLER_TEX0_SRGB] = srgb
    if invert:
        rs_invert_node = create_redshift_base_node(node_master, "RSMathInvColor", "RS Invert", posX+160, posY)
        rs_tex_node.GetOutPort(0).Connect(rs_invert_node.GetInPort(0))
        rs_invert_node.GetOutPort(0).Connect(rs_material.GetInPort(link))
        return rs_invert_node
    elif rs_material != None:
        rs_tex_node.GetOutPort(0).Connect(rs_material.GetInPort(link))
        return rs_tex_node

############# ARNOLD SETUP #########################
def create_arnold_shader(mat, ID, x, y, name = None, filename = None, clrspace = None):
    msg = c4d.BaseContainer()
    msg.SetInt32(1000, 1029)
    msg.SetInt32(2001, 1033990)
    msg.SetInt32(2002, ID)
    msg.SetInt32(2003, x)
    msg.SetInt32(2004, y)
    mat.Message(c4d.MSG_BASECONTAINER, msg)
    node = msg.GetLink(2011)
    if name:
        node.SetName(name)
    if filename:
        node.GetOpContainerInstance().SetFilename(1737748425, filename)
    if clrspace:
        node.GetOpContainerInstance().SetString(868305056, clrspace)
    return node

def arnold_set_base_shader(mat, shader, rootPortId):
    msg = c4d.BaseContainer()
    msg.SetInt32(1000, 1033)
    msg.SetLink(2001, shader)
    msg.SetInt32(2002, 0)
    msg.SetInt32(2003, rootPortId)
    mat.Message(c4d.MSG_BASECONTAINER, msg)
    return msg.GetBool(2011)

def arnold_mat_connection(mat, srcNode, dstNode, dstPortId):
    msg = c4d.BaseContainer()
    msg.SetInt32(1000, 1031)
    msg.SetLink(2001, srcNode)
    msg.SetInt32(2002, 0)
    msg.SetLink(2003, dstNode)
    msg.SetInt32(2004, dstPortId)
    mat.Message(c4d.MSG_BASECONTAINER, msg)
    return msg.GetBool(2011)

def arnold_material (texture_list, available_textures, matName, has_geometry, isMetal):
    doc = c4d.documents.GetActiveDocument()
    try:
        mat = c4d.BaseMaterial(1033991)
    except Exception as e:
        error_msg = "Message: Failed to create the material. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
        show_error_message(error_msg)
        print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)
        return False
    if mat is None:
        return False
    
    mat.SetName(matName)
    doc.InsertMaterial(mat)

    # Create master node and set it's properties.
    standard = create_arnold_shader(mat, 314733630, 500, 100, matName)
    arnold_set_base_shader(mat, standard, 537905099) # Connect to Arnold beauty output.
    arnold_set_base_shader(mat, standard, 537906863) # Connect to Arnold  viewport.
    standard.SetName(matName)
    standard.GetOpContainerInstance().SetString(1630484996, "custom")
    standard.GetOpContainerInstance().SetFloat(1182964519, 1) # Base: Weight
    standard.GetOpContainerInstance().SetFloat(1876347704, 0) # Specular Roughness Value
    standard.GetOpContainerInstance().SetFloat(1046994997, 0) # Specular Weight Value
    standard.GetOpContainerInstance().SetFloat(220096084, 1.52) # Specular IOR
    standard.GetOpContainerInstance().SetFloat(1875191464, 1 if isMetal else 0) # Metalness value

    try:
        for item in texture_list:

            if(c4d.GetC4DVersion() >= 23000):
                item = (item[0].decode('utf-8'), item[1].decode('utf-8'), item[2].decode('utf-8'))

            if item[1] == "albedo":
                albedo_node = create_arnold_shader(mat, 262700200, 0, 50, "Albedo", item[2], "sRGB")
                arnold_mat_connection(mat, albedo_node, standard, 1044225467)
                if "ao" in available_textures:
                    for tex in texture_list:
                        if tex[1] == "ao":
                            ao_node = create_arnold_shader(mat, 262700200, -250, 50, "AO", tex[2], "linear")
                            arnold_mat_connection(mat, ao_node, albedo_node, 1426019114) #Multiply with Albedo
            elif item[1] == "metalness" :
                metalness_node = create_arnold_shader(mat, 262700200, 0, 150, "Metalness", item[2], "linear")
                arnold_mat_connection(mat, metalness_node, standard, 1875191464)
            elif item[1] == "normal":
                normal_node = create_arnold_shader(mat, 262700200, 0, 200, "Normal Map", item[2], "linear")
                arnold_normal_node = create_arnold_shader(mat, 1512478027, 250, 200, "Arnold Normal Node")
                arnold_mat_connection(mat, normal_node, arnold_normal_node, 2075543287)
                arnold_mat_connection(mat, arnold_normal_node, standard, 244376085)
            elif item[1] == "gloss" and "roughness" not in available_textures:
                gloss_node = create_arnold_shader(mat, 262700200, 0, 100, "Gloss", item[2], "linear")
                gloss_cc_node = create_arnold_shader(mat, 1211336267, 250, 100, "Invert Gloss")
                gloss_cc_node.GetOpContainerInstance().SetBool(1952672187, True)
                gloss_cc_node.GetOpContainerInstance().SetFloat(1582985598, 0) # Exposure
                arnold_mat_connection(mat, gloss_node, gloss_cc_node, 2023242509)
                arnold_mat_connection(mat, gloss_cc_node, standard, 2099493681)
            elif item[1] == "roughness":
                roughness_node = create_arnold_shader(mat, 262700200, 0, 100, "Roughness", item[2], "linear")
                arnold_mat_connection(mat, roughness_node, standard, 2099493681)
            elif item[1] == "displacement":
                disp_node = create_arnold_shader(mat, 262700200, 400, 350, "Displacement", item[2], "linear")
                arnold_nrm_disp_node = create_arnold_shader(mat, 1270483482, 630, 300)
                arnold_nrm_disp_node.GetOpContainerInstance().SetInt32(748850620, 1) # Displacement scale.
                arnold_nrm_disp_node.GetOpContainerInstance().SetFloat(1762879056, 0.5) # Displacement scalar zero value.
                arnold_mat_connection(mat, disp_node, arnold_nrm_disp_node, 276937581)
                arnold_set_base_shader(mat, arnold_nrm_disp_node, 537905100)
            elif item[1] == "opacity":
                opacity_node = create_arnold_shader(mat, 262700200, 0, 250, "Opacity", item[2], "linear")
                arnold_mat_connection(mat, opacity_node, standard, 784568645)
            elif item[1] == "translucency":
                standard.GetOpContainerInstance().SetFloat(639969601, 0.1)
                standard.GetOpContainerInstance().SetFloat(657869786, 0.4)
                standard.GetOpContainerInstance().SetFloat(110275456, 1)
                standard.GetOpContainerInstance().SetVector(276268506, c4d.Vector(1,1,1))
                translucency_node = create_arnold_shader(mat, 262700200, 0, 300, "Translucency (SSS)", item[2], "sRGB")
                arnold_mat_connection(mat, translucency_node, standard, 676401187)
            elif item[1] == "transmission":
                transmission_node = create_arnold_shader(mat, 262700200, 0, 350, "Transmission", item[2], "linear")
                arnold_mat_connection(mat, transmission_node, standard, 1053345482)
                
  
    except Exception as e:
        print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)
        return None

    mat.Update(True, True)
    return mat

def ApplyMaterialToGeometry (mat, imported_geo, applyToSelection, isOpaque, useDisplacement, current_geo_import):
    doc = c4d.documents.GetActiveDocument()
    # create a texturetag and insert the material if apply to selected is false and we dont have to apply the material to imported geometry.
    if not ((applyToSelection and len(doc.GetActiveObjects(c4d.GETACTIVEOBJECTFLAGS_CHILDREN)) > 0) or len(imported_geo) > 0):
        tag = c4d.BaseTag( c4d.Ttexture )
        tag[c4d.TEXTURETAG_MATERIAL] = mat
        tag[c4d.TEXTURETAG_PROJECTION] = 6
        return

    rEngine = getRenderEngine()
    current_geo_import.append(imported_geo)
    scene_geo = doc.GetActiveObjects(c4d.GETACTIVEOBJECTFLAGS_CHILDREN)

    try:
        if applyToSelection:
            
            for item in scene_geo:
                if item not in current_geo_import:
                    # Apply the material to the model
                    tag = item.GetTag(c4d.Ttexture)
                    if tag == None:
                        tag = c4d.TextureTag()
                        item.InsertTag(tag)
                    tag.SetMaterial(mat)
                    tag[c4d.TEXTURETAG_PROJECTION] = 6
                    AddTagsToGeometry(rEngine, item, useDisplacement, isOpaque, False)
                

    except Exception as e:
        error_msg = "Message: Failed to apply texture to the selected objects. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
        show_error_message(error_msg)
        print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)
        return None

    try:
        for item in imported_geo:
            # Apply the material to the model
            tag = item.GetTag(c4d.Ttexture)
            if tag == None:
                tag = c4d.TextureTag()
                item.InsertTag(tag)
            tag.SetMaterial(mat)
            tag[c4d.TEXTURETAG_PROJECTION] = 6
            AddTagsToGeometry(rEngine, item, useDisplacement, isOpaque, True)

    except Exception as e:
        error_msg = "Message: Failed to apply material to the imported geometry. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
        show_error_message(error_msg)
        print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)
        return None

# Apply tags to the geometry according to the renderer.
def AddTagsToGeometry(rEngine, item, useDisplacement, isOpaque, importedGeometry): 
    #Redshift
    if rEngine == 1036219:
        Rtag = c4d.BaseTag(1036222)
        if useDisplacement:
            Rtag[c4d.REDSHIFT_OBJECT_GEOMETRY_OVERRIDE] = True
            Rtag[c4d.REDSHIFT_OBJECT_GEOMETRY_SUBDIVISIONENABLED] = True
            Rtag[c4d.REDSHIFT_OBJECT_GEOMETRY_DISPLACEMENTENABLED] = True
            Rtag[c4d.REDSHIFT_OBJECT_GEOMETRY_SMOOTHSUBDIVISIONENABLED] = False
            Rtag[c4d.REDSHIFT_OBJECT_GEOMETRY_MAXDISPLACEMENT] = 1 if importedGeometry else 10
            Rtag[c4d.REDSHIFT_OBJECT_GEOMETRY_DISPLACEMENTSCALE] = 1 if importedGeometry else 10
            Rtag[c4d.REDSHIFT_OBJECT_GEOMETRY_OVERRIDE] = False
        item.InsertTag(Rtag)
    # Octane
    elif rEngine == 1029525:
        Otag = c4d.BaseTag(1029603)
        item.InsertTag(Otag)
        phong = item.GetTag(c4d.Tphong) # check for phong tag on object
        if phong is None:
            phong[c4d.PHONGTAG_PHONG_ANGLELIMIT] = False

    # Arnold
    elif rEngine == 1029988:
        Atag = c4d.BaseTag(1029989)
        Atag[363257074] = 1
        Atag[1150993202] = 1
        Atag[1039494868] = 1
        Atag[1635638890] = 0
        Atag[486485632] = False
        Atag[408131505] = isOpaque
        Atag[1039494868] = 1 if importedGeometry else 10
        Atag[1635638890] = 0.5 if importedGeometry else 5
        Atag[1150993202] = 3
        item.InsertTag(Atag)

# refresh the current view so that all new materials are visible in the material window.
def Refresh():
    c4d.EventAdd()

# Get the active render engine and set it as default in the dropdown
def getRenderEngine():
    doc = c4d.documents.GetActiveDocument()
    RenderEngine = doc.GetActiveRenderData()[c4d.RDATA_RENDERENGINE]
    if RenderEngine == 1023342: RenderEngine = 0 # Physical
    return RenderEngine

def AddGeometry(filePath):

    doc = c4d.documents.GetActiveDocument()
    existing_geo = doc.GetObjects()

    c4d.documents.MergeDocument(doc, filePath, c4d.SCENEFILTER_OBJECTS, None)

    current_obj = doc.GetObjects()

    for i in current_obj:
        if i not in existing_geo:
            if (i.GetName() == "Game_Master"):
                baseName = os.path.basename(filePath)
                i.SetName(os.path.splitext(baseName)[0])

############# OCTANE SETUP #########################
class OctaneSetup():
    def __init__(self, textureList, textureTypes, assetType, hasGeomtry, matName, isSpecularWorkflow, isMetal):
        self.mat = None
        self.connectToTransformNode = False
        self.texture_list = textureList
        self.available_textures = textureTypes
        self.hasGeometry = hasGeomtry
        self.isSpecularWorkflow = isSpecularWorkflow
        self.doc = c4d.documents.GetActiveDocument()
        self.isMetal = isMetal
        self.assetType = assetType
        self.matName = matName

    def CreateMaterial(self):
        try:
            self.CreateMasterMaterialNode()
            self.SetupTextures()
            self.mat.Update(True, True)
            self.doc.InsertMaterial(self.mat)
            return self.mat
        except Exception as e:
            error_msg = "Message: CreateMaterial : Failed to create the octane material. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
            show_error_message(error_msg)
            print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)
            return None

    def CreateMasterMaterialNode(self):
        try:
            self.mat = c4d.BaseMaterial(1029501)
            self.mat[c4d.OCT_MATERIAL_TYPE] = 2516
            self.mat.SetName(self.matName)
            self.mat.Update(True, True)
            self.doc.InsertMaterial(self.mat)

            if self.isMetal:
                self.mat[c4d.OCT_MATERIAL_INDEX] = 8
        except Exception as e:
            error_msg = "Message: CreateMasterMaterialNode :Failed to create the octane material. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
            show_error_message(error_msg)
            print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)
        
        # Set BRDF model to GGX, this wasn't avaliable in older Octane versions
        try:
            self.mat[c4d.OCT_MAT_BRDF_MODEL] = 2
        except:
            pass

    def SetupTextures(self):
        try:
            if self.assetType.lower() not in ["3d","3dplant"]:
                self.CreateTransformNode()

            if "albedo" in self.available_textures:
                albedoNode = self.CreateImageTextureNode("albedo", "Albedo", 0, 2.2)
                if albedoNode:
                    ccAlbedoNode = self.CreateCCNode(albedoNode)
                    if "ao" in self.available_textures and not self.isMetal:
                        aoNode = self.CreateImageTextureNode("ao", "AO")
                        if aoNode:
                            self.CreateMultiplyNode(ccAlbedoNode, aoNode, c4d.OCT_MATERIAL_DIFFUSE_LINK)
                    else:
                        self.mat[c4d.OCT_MATERIAL_DIFFUSE_LINK] = ccAlbedoNode
            
            if self.isSpecularWorkflow:
                if "specular" in self.available_textures:
                    self.CreateImageTextureNode("specular", "Specular", 0, 2.2, False, c4d.OCT_MATERIAL_SPECULAR_LINK)
                
                if "gloss" in self.available_textures:
                    glossNode = self.CreateImageTextureNode("gloss", "Gloss", 1, 1, True)
                    if glossNode:
                        ccGlossNode = self.CreateCCNode(glossNode)
                        self.mat[c4d.OCT_MATERIAL_ROUGHNESS_LINK] = ccGlossNode
                elif "roughness" in self.available_textures:
                    roughnessNode = self.CreateImageTextureNode("roughness", "Roughness")
                    if roughnessNode:
                        ccRoughnessNode = self.CreateCCNode(roughnessNode)
                        self.mat[c4d.OCT_MATERIAL_ROUGHNESS_LINK] = ccRoughnessNode
            else:
                if "metalness" in self.available_textures:
                    self.CreateImageTextureNode("metalness", "Metalness", 1, 1, False, c4d.OCT_MAT_SPECULAR_MAP_LINK)
                if "roughness" in self.available_textures:
                    roughnessNode = self.CreateImageTextureNode("roughness", "Roughness")
                    if roughnessNode:
                        ccRoughnessNode = self.CreateCCNode(roughnessNode)
                        self.mat[c4d.OCT_MATERIAL_ROUGHNESS_LINK] = ccRoughnessNode

                elif "gloss" in self.available_textures:
                    glossNode = self.CreateImageTextureNode("gloss", "Gloss", 1, 1, True)
                    if glossNode:
                        ccGlossNode = self.CreateCCNode(glossNode)
                        self.mat[c4d.OCT_MATERIAL_ROUGHNESS_LINK] = ccGlossNode
                
            if "bump" in self.available_textures:  
                self.CreateImageTextureNode("bump", "Bump", 1, 1, False, c4d.OCT_MATERIAL_BUMP_LINK)

            if "normal" in self.available_textures:  
                self.CreateImageTextureNode("normal", "Normal", 0, 1, False, c4d.OCT_MATERIAL_NORMAL_LINK)
            
            if "displacement" in self.available_textures:
                displacementNode = self.CreateDisplacementNode(c4d.OCT_MATERIAL_DISPLACEMENT_LINK)
                if displacementNode:
                    displacementSlotName = c4d.DISPLACEMENT_TEXTURE
                    displacementNode[displacementSlotName] = self.CreateImageTextureNode("displacement", "Displacement")

            if "opacity" in self.available_textures:  
                self.CreateImageTextureNode("opacity", "Opacity", 1, 1, False, c4d.OCT_MATERIAL_OPACITY_LINK)

            if "translucency" in self.available_textures:  
                self.CreateImageTextureNode("translucency", "Translucency", 0, 2.2, False, c4d.OCT_MATERIAL_TRANSMISSION_LINK)
                self.mat[c4d.UNIVMAT_TRANSMISSION_TYPE] = 1
            elif "transmission" in self.available_textures:  
                self.CreateImageTextureNode("transmission", "Transmission", 1, 1, False, c4d.OCT_MATERIAL_TRANSMISSION_LINK)
                self.mat[c4d.UNIVMAT_TRANSMISSION_TYPE] = 1
        except Exception as e:
            error_msg = "Message: SetupTextures : Failed to create the octane material. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
            show_error_message(error_msg)
            print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)

    def TurnOffAllMaterialInputs(self):
        try:
            #Disable all shaders since they are enabled by default.
            self.mat[c4d.OCT_MAT_USE_COLOR] = False
            self.mat[c4d.OCT_MAT_USE_SPECULAR] = False
            self.mat[c4d.OCT_MAT_USE_ROUGHNESS] = False
            self.mat[c4d.OCT_MAT_USE_ANISO] = False
            self.mat[c4d.OCT_MAT_USE_SHEEN] = False
            self.mat[c4d.OCT_MAT_USE_COATING] = False
            self.mat[c4d.OCT_MAT_USE_FILMWIDTH] = False
            self.mat[c4d.OCT_MAT_USE_BUMP] = False
            self.mat[c4d.OCT_MAT_USE_NORMAL] = False
            self.mat[c4d.OCT_MAT_USE_DISPLACEMENT] = False
            self.mat[c4d.OCT_MAT_USE_OPACITY] = False
            self.mat[c4d.OCT_MAT_USE_DISPERSION] = False
            self.mat[c4d.OCT_MAT_USE_TRANSMISSION] = False
            self.mat[c4d.OCT_MAT_USE_EMISSION] = False
            self.mat[c4d.OCT_MAT_USE_MEDIUM] = False
            self.mat[c4d.OCT_MAT_USE_FAKESHADOW] = False
            self.mat[c4d.OCT_MAT_SPECULAR_MAP_FLOAT] = 0
        except:
            pass

    def CreateImageTextureNode(self, textureType, nodeName, textureMode = 1, gamma = 1, invert = False, parentNode = None):
        try:
            texturePath = self.GetTexturePath(textureType)
            if texturePath and os.path.exists(texturePath):
                imageTextureNode = c4d.BaseList2D(1029508)
                imageTextureNode[c4d.IMAGETEXTURE_FILE] = texturePath
                imageTextureNode[c4d.IMAGETEXTURE_INVERT] = invert
                imageTextureNode[c4d.IMAGETEXTURE_GAMMA] = gamma
                imageTextureNode[c4d.IMAGETEXTURE_MODE] = textureMode # 1 = Float and 0 is Normal (Color)
                imageTextureNode.SetName(nodeName)
                
                if self.connectToTransformNode:
                    imageTextureNode[c4d.IMAGETEXTURE_TRANSFORM_LINK] = self.transformNode
                
                if parentNode:
                    self.mat[parentNode] = imageTextureNode
                self.mat.InsertShader(imageTextureNode)
            
                return imageTextureNode
            else:
                print("Quixel Bridge Plugin::Failed to find the " + textureType + " texture.")
                return None
        except Exception as e:
            error_msg = "Message: CreateImageTextureNode : Failed to create the octane material. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
            show_error_message(error_msg)
            print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)
            return None
    
    def CreateTransformNode(self):
        try:
            self.transformNode = c4d.BaseList2D(1030961)
            self.mat.InsertShader(self.transformNode)
            self.connectToTransformNode = True
        except Exception as e:
            error_msg = "Message: CreateTransformNode : Failed to create the octane material. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
            show_error_message(error_msg)
            print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)

    def CreateMultiplyNode(self, imageTexture1, imageTexture2, parentNode = None):
        try:
            multiplyNode = c4d.BaseList2D(1029516)
            multiplyNode[c4d.MULTIPLY_TEXTURE1] = imageTexture1
            multiplyNode[c4d.MULTIPLY_TEXTURE2] = imageTexture2
            
            if parentNode:
                self.mat[parentNode] = multiplyNode
            self.mat.InsertShader(multiplyNode)
            
            return multiplyNode
        except Exception as e:
            error_msg = "Message: CreateMultiplyNode : Failed to create the octane material. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
            show_error_message(error_msg)
            print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)
            return None

    def CreateCCNode(self, imageTexture):
        try:
            colorCorrectionNode = c4d.BaseList2D(1029512)
            colorCorrectionNode[c4d.COLORCOR_TEXTURE_LNK] = imageTexture
            self.mat.InsertShader(colorCorrectionNode)
            return colorCorrectionNode
        except Exception as e:
            error_msg = "Message: CreateCCNode : Failed to create the octane material. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
            show_error_message(error_msg)
            print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)
            return None

    def CreateDisplacementNode(self, parentNode = None):
        try:
            displacementNode = c4d.BaseList2D(1031901)
            displacementNode[c4d.DISPLACEMENT_AMOUNT] = 1 if self.assetType.lower() in ["3d","3dplant"] else 10
            displacementNode[c4d.DISPLACEMENT_MID] = 0.5
            self.mat[c4d.OCT_MAT_USE_DISPLACEMENT] = not self.hasGeometry
            
            if parentNode:
                self.mat[parentNode] = displacementNode
            self.mat.InsertShader(displacementNode)

            return displacementNode
        except Exception as e:
            error_msg = "Message: CreateDisplacementNode : Failed to create the octane material. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
            show_error_message(error_msg)
            print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)
            return None

    def GetTexturePath(self, textureType):
        try:
            if(c4d.GetC4DVersion() < 23000):
                for item in self.texture_list:
                    if item[1] == textureType:
                        return item[2].replace("\\", "/")
            else:
                for item in self.texture_list:
                    if item[1] == textureType.encode('utf-8'):
                        return item[2].decode('utf-8').replace("\\", "/")
        except Exception as e:
            error_msg = "Message: GetTexturePath : Failed to create the octane material. \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
            show_error_message(error_msg)
            print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)
            return None

# Creation Part
class CreationPart(c4d.plugins.MessageData):
    
    def CoreMessage(self, id, msg):
        if id == pluginID:
            # Make sure nothing is written the scene currently, it will not stop our custom thread.
            c4d.StopAllThreads()

            # Get the Void Container of the message
            # P1MSG_UN = msg.GetVoid(c4d.BFM_CORE_PAR1)
            
            # # Convert PyObject to integer (retrieve the value passed)
            # pythonapi.PyCapsule_GetPointer.restype = c_void_p
            # pythonapi.PyCapsule_GetPointer.argtypes = [py_object]
            index = 0#int(pythonapi.PyCObject_AsVoidPtr(P1MSG_UN))

            # Get the worldContainer with our Data
            bcWorld = c4d.GetWorldContainerInstance()

            # Get the basecontainer which contain all our settings
            bcSetting = bcWorld.GetContainerInstance(pluginID)

            # Get the basecontainer which contains all our Message
            bcPlugin = bcWorld.GetContainerInstance(BC_WORLD_ID)
            if not bcPlugin: return

            # Get the data stored in the index passed as argument
            bcData = bcPlugin[index]
            if not bcData: return

            # Then according of the ID of the Basecontainer we do different things
            if bcData.GetId() == BC_IMPORT_MS_ASSET:
                print("Quixel Bridge Plugin::Received data from Bridge.")
                if(bcSetting != None):
                    MS_Importer(bcData[0], bcSetting[0], bcSetting[1], bcSetting[2])
                else:
                    MS_Importer(bcData[0], False, False, False)
            else:
                print ("Quixel Bridge Plugin::Operation cannot be recognized.")

            # Our data is processed we clean our bc
            bcPlugin.RemoveData(index)

            # Inform Cinema 4D to trigger an update
            c4d.EventAdd()

        return True

# The Diaog class
class DialogSetting(c4d.gui.GeDialog):
    
    def InitValues(self):
        # Get the worldContainer with our Data
        bcWorld = c4d.GetWorldContainerInstance()

                # Get the basecontainer which contain settings
        bcSetting = bcWorld.GetContainerInstance(pluginID)
        if not bcSetting: return True
        
        # Read values stored in the BaseContainer
        self.SetBool(1001, bcSetting[0])
        self.SetBool(1002, bcSetting[1])
        self.SetBool(1003, bcSetting[2])

        return True

    # Define our Layout
    def CreateLayout(self):

        self.SetTitle("Megascans Plugin v" + pluginVersion)

        self.GroupBegin(id=1, flags=c4d.BFH_SCALEFIT, rows=3, title=("Megascans Plugin v" + pluginVersion), cols=1, initw = 550)
        # self.AddStaticText(id=2, flags=c4d.BFH_CENTER, initw=300, name="Megascans Plugin v" + pluginVersion)
        # self.AddSeparatorH(inith=1)
        self.AddCheckbox(1001, c4d.BFH_CENTER, 300, 10, "Use EXRs when available")
        self.AddCheckbox(1002, c4d.BFH_CENTER, 300, 10, "Apply Material to Selection")
        self.AddCheckbox(1003, c4d.BFH_CENTER, 300, 10, "Enable Displacement")
        self.GroupEnd()
    
        return True

    def Command(self, id, msg):
        # If user click on any checkbox, we save the result in 

        if id in [1001, 1002, 1003]:
            # Get the worldContainer with our Data
            bcWorld = c4d.GetWorldContainerInstance()

            # Get the basecontainer which contain settings
            bcSetting = bcWorld.GetContainerInstance(pluginID)
            if not bcSetting:
                bcSetting = c4d.BaseContainer()
                bcWorld[pluginID] = bcSetting
                bcSetting = bcWorld.GetContainerInstance(pluginID)

            # Store our datas in the world container
            bcSetting[0] = self.GetBool(1001)
            bcSetting[1] = self.GetBool(1002)
            bcSetting[2] = self.GetBool(1003)
            
            return True

# The CommandData plugin which host the Dialog object alive
class CommandDataDlg(c4d.plugins.CommandData):
        dialog = None

        def Execute(self, doc):
            if self.dialog is None:
                self.dialog = DialogSetting()
            return self.dialog.Open(dlgtype=c4d.DLG_TYPE_ASYNC, pluginid=pluginIDCommandData, xpos=-1, ypos=-1, defaultw=200, defaulth=150)

        def RestoreLayout(self, sec_ref):
        # manage the dialog
            if self.dialog is None:
                self.dialog = DialogSetting()
            return self.dialog.Restore(pluginid=pluginIDCommandData, secret=sec_ref)

# Handle our thread here
def PluginMessage(id, data):
    global thread_
    
    # At the start of Cinema 4D We lunch our thread
    if id == c4d.C4DPL_PROGRAM_STARTED:
        thread_ = BGThread()
        thread_.Start()

    # At the end we clean our Thread + BC used to pass data
    elif id == c4d.C4DPL_ENDACTIVITY:
       
        try:
            s = socket.socket()
            s.connect((host,port))
            data = "Bye Megascans"
            s.send(data.encode())
            s.close()
        except Exception as e:
            error_msg = "Message: Failed to start the Plugin \nError: " + type(e).__name__ + "\nLine no: " + str(sys.exc_info()[-1].tb_lineno) + "\nDescription: " + str(e) 
            show_error_message(error_msg)
            print('Quixel Bridge Plugin::Error Line : {}'.format(sys.exc_info()[-1].tb_lineno), type(e).__name__, e)

        if thread_:
            thread_.end = True
            thread_.End()

        # Clean our global Container
        bcWorld = c4d.GetWorldContainerInstance()
        bcWorld.RemoveData(BC_WORLD_ID)

def show_error_message (msg):
    finalMessage = "Bridge plugin " + pluginVersion + "\n" + msg
    gui.MessageDialog(finalMessage)

# Main function
def main():
    icon = c4d.bitmaps.BaseBitmap()
    ScriptPath = os.path.split(__file__)[0]
    iconPath = os.path.join(ScriptPath, "megascans_icon.jpg")
    icon.InitWith(iconPath)
    c4d.plugins.RegisterMessagePlugin(pluginID, "Data parser and importer.", 0, CreationPart())
    c4d.plugins.RegisterCommandPlugin(id=pluginIDCommandData, str="Megascans Plugin", help="Megascans Birdge - Cinema4D Plugin", info=0, dat=CommandDataDlg(), icon=icon)

# Execute main()
if __name__=='__main__':
    main()